/*
** Apple Macintosh Developer Technical Support
**
** File:		DoEvent.c
** Written by:	Eric Soldan
**
** Copyright  1990-1993 Apple Computer, Inc.
** All rights reserved.
*/

/* You may incorporate this sample code into your applications without
** restriction, though the sample code has been provided "AS IS" and the
** responsibility for its operation is 100% yours.  However, what you are
** not permitted to do is to redistribute the source as "DSC Sample Code"
** after having made changes. If you're going to re-distribute the source,
** we require that you make it clear in the source that the code was
** descended from Apple Sample Code, but that you've made changes. */



/*****************************************************************************/



#include "App.h"			/* Get the application includes/typedefs, etc.	*/
#include "App.protos.h"		/* Get the prototypes for application.			*/

#ifndef __CTLHANDLER__
#include "CtlHandler.h"
#endif

#ifndef __DESK__
#include <Desk.h>
#endif

#ifndef __DISKINIT__
#include <DiskInit.h>
#endif

#ifndef __ERRORS__
#include <Errors.h>
#endif

#ifndef __MENUS__
#include <Menus.h>
#endif

#ifndef __TEXTEDITCONTROL__
#include "TextEditControl.h"
#endif

#ifndef __TOOLUTILS__
#include <ToolUtils.h>
#endif

#ifndef __UTILITIES__
#include "Utilities.h"
#endif



/*****************************************************************************/



extern Cursor	*gCursorPtr;	/* See the file Window.c for comments about this global. */



/*****************************************************************************/
/*****************************************************************************/



/* Do the right thing for an event.  Determine what kind of event it is, and
** call the appropriate routines. */

#pragma segment Main
void	DoEvent(EventRecord *event)
{
	WindowPtr	window;
	Point		pt;
	OSErr		err;
	TEHandle	teHndl;
	Rect		rct;

	switch(event->what) {

		case nullEvent:
			DoIdleTasks(event);
			break;

		case mouseDown:
			DoMouseDown(event);			/* Let framework handle mouse downs. */
			break;

		case autoKey:
		case keyDown:
			DoKeyDown(event);			/* Let framework handle key events. */
			break;

		case activateEvt:
			gCursorPtr = nil;			/* Force recalculation of the cursor. */
			DoActivate((WindowPtr)event->message);
			break;

		case updateEvt:
			DoUpdate((WindowPtr)event->message);
			break;

		case kHighLevelEvent:
			gCursorPtr = nil;			/* Force recalculation of the cursor. */
			DoHighLevelEvent(event);
			break;

		case osEvt:
			gCursorPtr = nil;			/* Force recalculation of the cursor. */
			switch ((event->message >> 24) & 0xFF) {
					/* Must logical and with 0xFF to get only low byte. */
					/* High byte of message. */

				case mouseMovedMessage:
					DoIdleTasks(event);		/* Treat mouseMoved events like NULL events. */
					break;

				case suspendResumeMessage:
						/* Suspend/resume is also an activate/deactivate. */
					gInBackground = !((event->message) & resumeFlag);
					CTEConvertClipboard((event->message) & convertClipboardFlag, !gInBackground);
					DoActivate(FrontWindow());
					HiliteWindows();
					break;
			}
			break;

		case diskEvt:
			gCursorPtr = nil;			/* Force recalculation of the cursor. */
			if (HiWord(event->message) != noErr) {
				SetPt(&pt, kDILeft, kDITop);
				err = DIBadMount(pt, event->message);
			}		/* It is not a bad idea to at least call DIBadMount in response to */
			break;	/* a diskEvt, so that the user can format a floppy. */
	}

	DoCursor();
	DoAdjustMenus();

	if (window = CTETargetInfo(&teHndl, &rct)) {
		if (rct.left < -8192)		/* If TextEdit control is in the frame...  */
			BeginFrame(window);		/* Set clipping to the frame area.		   */
		else
			BeginContent(window);	/* Else set clipping to the document area. */
		CTEIdle();
		EndContent(window);			/* EndContent can be used to close a BeginFrame. */
	}
}



/*****************************************************************************/



/*  Called by DTS.Lib..framework.  */

/* This is called when a window is activated or deactivated. */

#pragma segment Main
void	DoActivate(WindowPtr window)
{
	NotifyCancel();

	if (IsAppWindow(window)) {
		SetPort(window);
		DoCtlActivate(window);
		DoDrawFrame(window, true);			/* Redraw frame for new activate state. */
		BeginContent(window);
		DoDrawControls(window, true);		/* Redraw content scrollbars for new activate state. */
		EndContent(window);
	}
}



